import sys
import random
import copy
import numpy
import scipy.stats
import math
import pdb
import cProfile
import cPickle
import optparse
import re
from operator import itemgetter
import multiprocessing
import logging
import collections
import warnings
import ConfigParser
import time
import string

import sb08_data_helper

class FastDirichletMult:
  def __init__(self,alpha,dim):
    assert alpha >= 0
    assert dim > 0
    self.alpha = float(alpha)
    self.dim = dim
    self.counts = [0] * dim
    self.total = 0

  def validate(self):
    return all([v >= 0 for v in self.counts]) \
           and self.total == sum(self.counts)

  def __eq__(self,other):
    return (self.alpha == other.alpha) \
           and (self.dim == other.dim) \
           and (self.counts == other.counts) \
           and (self.total == other.total)

  def __ne__(self,other): return not self.__eq__(other)
  
  def observe(self, k, v):
    assert k >= 0
    assert type(v) == int
    self.counts[k] += v
    self.total += v

  def log_prob(self, k):
    assert k >= 0
    return math.log( (self.alpha + self.counts[k]) / (self.alpha * self.dim + self.total) )

  def log_prob_with_N(self, k, N):
    assert k >= 0
    return math.log( (self.alpha + self.counts[k]) / (self.alpha * N + self.total) )

class DirichletMult:
  def __init__(self,alpha):
    assert alpha > 0
    self.alpha = float(alpha)
    self.counts = {}
    self.total = 0

  def __eq__(self, other):
    alpha_eq = (self.alpha == other.alpha)
    counts_eq = (self.counts == other.counts)
    total_eq = (self.total == other.total)
    all_eq = alpha_eq and counts_eq and total_eq
    return all_eq

  def __ne__(self, other):
    return not self.__eq__(other)

  def __str__(self):
    n = sum(self.counts.values())
    p = sorted([ (k,"%5.4f" % (1.0*v/n)) for k,v in self.counts.iteritems() ], key=lambda x: x[1], reverse=True)
    return str(p)

  def observe(self, k, v):
    assert type(v) == int
    if k not in self.counts:
      self.counts[k] = v
    else:
      self.counts[k] = self.counts[k] + v
    self.total = self.total + v
    assert self.counts[k] >= 0 # TODO
    if self.counts[k] == 0:
      del self.counts[k]

  def vocab_size(self):
    n = len(self.counts)
    return n

  def prob(self, k, N):
    numer = self.alpha + self.counts[k] if k in self.counts \
            else self.alpha
    assert 0 <= self.vocab_size() <= N
    assert self.total >= 0
    denom = self.alpha * N + self.total
    return numer / denom

  def log_prob(self, k, N):
    return math.log(self.prob(k,N))

class FastDirichletMultArray:
  def __init__(self, alpha, num_distrs):
    assert alpha >= 0
    self.alpha = float(alpha)
    self.counts = {}
    self.totals = [0] * num_distrs
    self.nd = num_distrs

  def __eq__(self, other):
    return (self.alpha == other.alpha) \
           and (self.counts == other.counts) \
           and (self.totals == other.totals) \
           and (self.nd == other.nd)

  def __ne__(self, other): return not self.__eq__(other)

  def observe(self, i, k, v):
    assert type(v) == int
    assert i >= 0
    if k not in self.counts: self.counts[k] = [0] * self.nd
    c = self.counts[k]
    c[i] += v
    self.totals[i] += v
    if c[i] == 0 and sum(c) == 0:
      del self.counts[k]

  def log_probs_no_tag(self, k, N):
    assert self.totals[0] == 0
    sum_totals = sum(self.totals)
    if k in self.counts:
      countsk = self.counts[k]
      assert countsk[0] == 0
      sum_counts = sum(countsk)
      return math.log( (self.alpha + sum_counts) / (self.alpha*N + sum_totals) )
    else:
      return math.log( (self.alpha / (self.alpha*N + sum_totals) ) )
    
  def log_probs(self, k, N, tags):
    if k in self.counts:
      countsk = self.counts[k]
      return [ math.log( (self.alpha + countsk[i]) / (self.alpha*N + self.totals[i]) ) \
               for i in tags ]
    else:
      return [ math.log( self.alpha / (self.alpha*N + self.totals[i]) ) \
               for i in tags ]

  def log_probs_with_Ns(self, k, Ns, tags):
    assert len(Ns) == len(tags)
    if k in self.counts:
      countsk = self.counts[k]
      return [ math.log( (self.alpha + countsk[i]) / (self.alpha*N + self.totals[i]) ) \
               for (N,i) in zip(Ns,tags) ]
    else:
      return [ math.log( self.alpha / (self.alpha*N + self.totals[i]) ) \
               for (N,i) in zip(Ns,tags) ]

  def validate(self):
    return self.alpha >= 0 \
           and all([ x >= 0 for x in self.counts.values() ]) \
           and all([ y >= 0 for y in self.totals ]) \
           and (self.nd == len(self.totals))
    
class LexiconState:
  RESTRICT_SEG_COUNTS = True
  RESTRICT_TYPE_TAG_COUNTS = True
  BOUNDARY_TOKEN = "~~"
  MAX_SPANS = 5
  MAX_AFFIX = 2
  MODEL_STEM_POSITION = True
  ALPHA_TAG_PRIOR = 0.1
  ALPHA_EMISSION = 1e-5 # 0.1
  ALPHA_TRANSITION = 1.0 # 0.1
  CONFLICT_STEM_PROB = 0.01
  CONFLICT_LOG_ODDS = math.log( CONFLICT_STEM_PROB / (1.0 - CONFLICT_STEM_PROB) )
  FROZEN_RE = re.compile("^[\d%s]+$" % re.escape(string.punctuation))

  def __init__(self, hyperparameters, word_freq,seq_data,gold,tag_lexicon, init_tag, init_seg, init_stem, num_tags, state0, sep_lex_size, use_seq_suffix, use_seq_prefix):
    self.ALPHA_SEG = 0.1                   if 'ALPHA_SEG' not in hyperparameters else hyperparameters['ALPHA_SEG']
    self.GAMMA_NUM_UNIQUE_SEGS = 1.0/10    if 'GAMMA_NUM_UNIQUE_SEGS' not in hyperparameters else hyperparameters['GAMMA_NUM_UNIQUE_SEGS']
    self.GAMMA_SEG_LEN = 1.0/2.0           if 'GAMMA_SEG_LEN' not in hyperparameters else hyperparameters['GAMMA_SEG_LEN']
    self.GAMMA_NUM_SEGS_PER_WORD = 1.0/4.0 if 'GAMMA_NUM_SEGS_PER_WORD' not in hyperparameters else hyperparameters['GAMMA_NUM_SEGS_PER_WORD']
    self.GAMMA_AFFIX_PER_WORD = 1.0/2      if 'GAMMA_AFFIX_PER_WORD' not in hyperparameters else hyperparameters['GAMMA_AFFIX_PER_WORD']
    self.NUM_TAGS = num_tags
    self.VALENCE_MODEL = 0
    self.LOG_GEOM_TRUNC_NUM_SEGS_PER_WORD = LogGeometricTruncated(self.GAMMA_NUM_SEGS_PER_WORD,
                                                                  LexiconState.MAX_SPANS)
    self.LOG_GEOM_TRUNC_NUM_AFFIX_PER_WORD = LogGeometricTruncated(self.GAMMA_AFFIX_PER_WORD,
                                                                   LexiconState.MAX_AFFIX)
    self.words = {}
    self.USE_PAIRWISE_SUFFIXES = use_seq_suffix
    self.USE_PAIRWISE_PREFIXES = use_seq_prefix
    self.PAIRWISE_SUFFIX_LOGPROB = { None: math.log(0.3), True: math.log(0.6), False: math.log(0.1) }
    self.PAIRWISE_PREFIX_LOGPROB = { None: math.log(0.05), True: math.log(0.9), False: math.log(0.05) }

    self.USE_STEM_LENGTH_CONSTRAINT = True
    self.SEPARATE_LEXICON_SIZE = sep_lex_size
    self.X_GAMMA_NUM_UNIQUE_SEGS = {}
    self.X_GAMMA_NUM_UNIQUE_SEGS['stem']   = 1.0/10000 if 'X_GAMMA_NUM_UNIQUE_SEGS-STEM'   not in hyperparameters else hyperparameters['X_GAMMA_NUM_UNIQUE_SEGS-STEM']
    self.X_GAMMA_NUM_UNIQUE_SEGS['prefix'] = 1.0/1.1   if 'X_GAMMA_NUM_UNIQUE_SEGS-PREFIX' not in hyperparameters else hyperparameters['X_GAMMA_NUM_UNIQUE_SEGS-PREFIX']
    self.X_GAMMA_NUM_UNIQUE_SEGS['suffix'] = 1.0/1.1   if 'X_GAMMA_NUM_UNIQUE_SEGS-SUFFIX' not in hyperparameters else hyperparameters['X_GAMMA_NUM_UNIQUE_SEGS-SUFFIX']

    if init_tag == "post" or init_tag == "gold":
      assert tag_lexicon

    if init_seg == "gold":
      assert False

    if init_stem == "gold":
      assert init_seg == "gold"
      # TODO
      assert False

    for word,freq in word_freq:
      assert word
      assert word != LexiconState.BOUNDARY_TOKEN
      assert freq >= 0
      to_freeze = LexiconState.FROZEN_RE.match(word)
      to_segment = not to_freeze
      to_stem = True
      to_tag = True
      
      # valid tag in [1,NUM_TAGS]
      # tag = 0 used for start,stop states
      tag = None # in case assertions are disabled
      if init_tag == "state":
        tag = state0.words[word].tag
      elif init_tag == "uniform":
        tag = 1 if self.NUM_TAGS == 1 else random.randint(1,self.NUM_TAGS)
      else: assert False

      spans = None
      if init_seg == "whole-word":
        spans = [ (0,len(word)) ] # whole word
      elif init_seg == "state":
        spans = copy.deepcopy(state0.words[word].spans)
      else: assert False

      stem_index = None
      if init_stem == "state":
        assert init_seg == "state"
        stem_index = state0.words[word].stem_index
      elif init_stem == "first":
        stem_index = 0
      else: assert False

      ws = WordState(word,freq,tag, tuple(spans), stem_index, to_segment, to_stem, to_tag)
      self.words[word] = ws
    # for
    (self.counter_seg,
     self.x_counter_seg,
     self.x_fast_distrs_seg_gt,
     self.distr_type_tag
     ) = self.get_counts()

    self.tokens = seq_data
    if seq_data:
      self.word_locations = LexiconState.get_word_locations(seq_data)
      (self.fast_distr_token_gt,self.distrs_transition) = self.get_seq_token_counts()
          
  @staticmethod
  def get_boundaries(segs, n, pad):
    b = []
    for i,seg in enumerate(segs):
      if i == 0: continue
      right = seg[:n].ljust(n,pad)
      left = segs[i-1][-n:].rjust(n,pad)
      b.append( (left,right) )
    return b

  def get_seq_token_counts(self):
    fast_distrs_token_gt = FastDirichletMultArray(LexiconState.ALPHA_EMISSION, self.NUM_TAGS+1)
    distrs_transition = [ FastDirichletMult(LexiconState.ALPHA_TRANSITION, self.NUM_TAGS+1)
                          for t in range(self.NUM_TAGS+1) ]
    for i,w in enumerate(self.tokens):
      if i == 0: continue
      t = self.words[w].tag if w != LexiconState.BOUNDARY_TOKEN else 0

      if w != LexiconState.BOUNDARY_TOKEN:
        fast_distrs_token_gt.observe(t, w, 1)

      w_ = self.tokens[i-1]
      t_ = self.words[w_].tag if w_ != LexiconState.BOUNDARY_TOKEN else 0
      distrs_transition[t_].observe(t, 1)
    return (fast_distrs_token_gt,distrs_transition)

  def get_counts(self):
    # type-level counts
    distr_type_tag = DirichletMult(LexiconState.ALPHA_TAG_PRIOR)
    counter_seg = collections.Counter()
    x_counter_seg = {}
    x_fast_distrs_seg_gt = {}
    for affix in WordState.POSSIBLE_AFFIXES:
      x_counter_seg[affix] = collections.Counter()
      x_fast_distrs_seg_gt[affix] = FastDirichletMultArray(self.ALPHA_SEG, self.NUM_TAGS+1)

    for w,ws in self.words.iteritems():
      t = ws.tag
      if not LexiconState.RESTRICT_TYPE_TAG_COUNTS or \
             (LexiconState.RESTRICT_TYPE_TAG_COUNTS and ws.to_tag):
        distr_type_tag.observe(t,1)
      if not LexiconState.RESTRICT_SEG_COUNTS or \
             (LexiconState.RESTRICT_SEG_COUNTS and ws.to_stem):
        for i,s in enumerate(ws.segments()):
          counter_seg[s] += 1
          affix = ws.seg_affix(i)
          x_fast_distrs_seg_gt[affix].observe(t, s, 1)
          x_counter_seg[affix][s] += 1

    # for
    return (counter_seg,
            x_counter_seg,
            x_fast_distrs_seg_gt,
            distr_type_tag)

  @staticmethod
  def get_word_locations(tokens):
    d = {}
    for i,t in enumerate(tokens):
      assert t.find(' ') < 0 # assert not a phrase
      if True or t != LexiconState.BOUNDARY_TOKEN:
        if t not in d: d[t] = []
        d[t].append(i)
    e = {}
    for t,locs in d.iteritems():
      e[t] = tuple(locs)
    return e

  def validate_num_spans(self):
    num_spans_within_limits = True
    for w,ws in self.words.iteritems():
      if len(ws.spans) > LexiconState.MAX_SPANS:
        num_spans_within_limits = False
        break
    return num_spans_within_limits

  def validate_tag_range(self):
    tag_within_range = True
    for w,ws in self.words.iteritems():
      if not 1 <= ws.tag <= self.NUM_TAGS:
        tag_within_range = False
        break
    return tag_within_range

  def observe_word_state(self, ws, count):
    assert type(count) == int
    t = ws.tag
    w = ws.word
    if not LexiconState.RESTRICT_TYPE_TAG_COUNTS or \
           (LexiconState.RESTRICT_TYPE_TAG_COUNTS and ws.to_tag):
      self.distr_type_tag.observe(t, count)

    if not LexiconState.RESTRICT_SEG_COUNTS or \
           (LexiconState.RESTRICT_SEG_COUNTS and ws.to_stem):
      for i,s in enumerate(ws.segments()):
        self.counter_seg[s] += count
        if self.counter_seg[s] == 0: del self.counter_seg[s]

        affix = ws.seg_affix(i)
        self.x_fast_distrs_seg_gt[ affix ] . observe(t, s, count)
        c = self.x_counter_seg[affix]
        c[s] += count
        if c[s] == 0: del c[s]
    # if
    if self.tokens:
      for i in self.word_locations[w]:
        assert self.tokens[i] == w
        assert w != LexiconState.BOUNDARY_TOKEN
        self.fast_distr_token_gt.observe(t, w, count)
        w_ = self.tokens[i-1]
        if w_ == LexiconState.BOUNDARY_TOKEN:
          t_ = 0
        elif w_ == w:
          t_ = t # w not added yet
        else:
          t_ = self.words[w_].tag

        self.distrs_transition[t_].observe(t, count)
        _w = self.tokens[i+1]
        if _w != w:
          if _w == LexiconState.BOUNDARY_TOKEN:
            _t = 0
          elif _w == w:
            _t = t # w not added yet
          else:
            _t = self.words[_w].tag
          self.distrs_transition[t].observe(_t, count)
    # if self.tokens

  def add_word_state(self, ws):
    w = ws.word
    assert w not in self.words
    self.observe_word_state(ws, 1)
    self.words[w] = copy.deepcopy(ws)

  def remove_word(self, w):
    ws = self.words[w]
    self.observe_word_state(ws, -1)
    del self.words[w]
    return ws

  def to_segmented_lexicon(self):
    d = {}
    for w,ws in self.words.iteritems():
      s = ws.segments()
      d[w] = s
    return d

  def num_prefixes_per_word(self):
    return [ ws.stem_index
             for ws in self.words.values() if ws.to_segment ]
    
  def num_suffixes_per_word(self):
    return [ len(ws.spans) - ws.stem_index - 1
             for ws in self.words.values() if ws.to_segment ]

  def seg_lens(self):
    return [ len(s)
             for ws in self.words.values()
             for s in ws.segments() if ws.to_segment ]

  def normalized_num_seg(self):
    a = [ 1.0*len(ws.spans)/len(w) for w,ws in self.words.iteritems() ]
    #(h,b) = numpy.histogram(a)
    #for i,j in zip(h,b): print i,j
    return a

  def num_seg(self):
    a = [ len(ws.spans) for w,ws in self.words.iteritems() ]
    return a

  def prefixes(self):
    return [ w[start:stop]
             for w,ws in self.words.iteritems() if ws.to_segment
             for start,stop in ws.spans[ 0 : ws.stem_index] ]
  
  def num_prefix_types(self):
    return len(set(self.prefixes()))

  def suffixes(self):
    return [ w[start:stop]
             for w,ws in self.words.iteritems() if ws.to_segment
             for start,stop in ws.spans[ ws.stem_index+1 :: ] ]
    
  def num_suffix_types(self):
    return len(set(self.suffixes()))

  def stems(self):
    return [ w[ ws.spans[ws.stem_index][0] :
                ws.spans[ws.stem_index][1] ]
             for w,ws in self.words.iteritems() if ws.to_segment ]

  def get_affix_segment_lengths(self):
    d = collections.defaultdict(list)
    
    for w,ws in self.words.iteritems():
      if ws.to_segment:
        for i,(start,stop) in enumerate(ws.spans):
          affix = ws.seg_affix(i)
          d[affix].append(stop-start)
    return d

  def get_min_affix_vocab_size(self, segs, stem_index):
    affixes = {}
    for i,s in enumerate(segs):
      x = WordState.get_seg_affix(i, stem_index)
      if x not in affixes: affixes[x] = set()
      affixes[x].add(s)
      
    sz = {}
    for x,s in affixes.iteritems():
      c = self.x_counter_seg[x]
      current_size = len(c)
      num_new_types = len([seg for seg in s if c[seg] == 0])
      sz[x] = current_size + num_new_types
      
    return sz

  def validate(self):
    ok_counter_seg =  not [v for v in self.counter_seg.values() if v <= 0]
    ok_x_counter_seg = reduce(lambda retval,c: retval and not [v for v in c.values() if v <= 0], \
                              self.x_counter_seg.values(), True)
    ok_x_fast_distrs_seg_gt = reduce(lambda retval, d: retval and d.validate(), self.x_fast_distrs_seg_gt.values(), True)

    (counter_seg,
     x_counter_seg,
     x_fast_distrs_seg_gt,
     distr_type_tag
     ) = self.get_counts()
    
    match_counter_seg = (counter_seg == self.counter_seg)
    match_x_counter_seg = (x_counter_seg == self.x_counter_seg)
    match_x_fast_distrs_seg_gt = (x_fast_distrs_seg_gt == self.x_fast_distrs_seg_gt)
    match_distr_type_tag = (distr_type_tag == self. distr_type_tag)

    if self.tokens:
      (fast_distr_token_gt,distrs_transition) = self.get_seq_token_counts()
      fast_distr_token_gt_ok = fast_distr_token_gt == self.fast_distr_token_gt
      distrs_transition_ok = distrs_transition == self.distrs_transition
      no_null_trans = self.distrs_transition[0].counts[0] == 0
      seq_ok = fast_distr_token_gt_ok and distrs_transition_ok and no_null_trans
    else:
      seq_ok = True

    all_ok = ok_counter_seg \
           and ok_x_counter_seg \
           and ok_x_fast_distrs_seg_gt \
           and match_counter_seg \
           and match_x_counter_seg \
           and match_x_fast_distrs_seg_gt \
           and match_distr_type_tag
    if not all_ok: pdb.set_trace()
    return ok_counter_seg \
           and ok_x_counter_seg \
           and ok_x_fast_distrs_seg_gt \
           and match_counter_seg \
           and match_x_counter_seg \
           and match_x_fast_distrs_seg_gt \
           and match_distr_type_tag

  def log_prob_segment_length(self,k):
    return log_geometric(self.GAMMA_SEG_LEN, k-1)

  def log_prob_num_segments_in_word(self, word_length, num_segs):
    #k = 1.0 * num_segs
    #return log_geometric(LexiconState.GAMMA_NORMALIZED_NUM_SEGS_PER_WORD, k)
    return log_geometric_truncated(self.GAMMA_NUM_SEGS_PER_WORD, num_segs, LexiconState.MAX_SPANS)

  def get_log_tag_prior(self, all_possible_tags):
    return [ self.distr_type_tag.log_prob(t, self.NUM_TAGS)
             for t in all_possible_tags ]
##     K = self.NUM_TAGS
##     p = [ -float('inf') ] * (K+1)
##     for t in all_possible_tags:
##       p[t] = self.distr_type_tag.log_prob(t, K)
##     return p
    
  def log_token_emission(self,w,tags):
    d = self.fast_distr_token_gt
    assert w != LexiconState.BOUNDARY_TOKEN
    assert w not in d.counts

    nw = len(self.word_locations[w])
    V = len(d.counts)
    log_probs = [0.0] * len(tags)
    for j,t in enumerate(tags):
      total = d.totals[t]
      for i in range(nw):
        numer = d.alpha + i
        denom = d.alpha * V + total + i 
        log_probs[j] += math.log( numer / denom )
    return log_probs
  
  def log_token_trans(self, w,t):
    def increment_transition_counts(d, t_, t):
      if t_ == 0: assert t != 0
      if t_ not in d: d[t_] = {}
      if t not in d[t_]:
        d[t_][t] = 1
      else:
        d[t_][t] += 1
      
    assert w != LexiconState.BOUNDARY_TOKEN
    tc = {}
    for i in self.word_locations[w]:
      assert self.tokens[i] == w
      w_ = self.tokens[i-1]

      if w_ == LexiconState.BOUNDARY_TOKEN:
        t_ = 0
      elif w_ == w:
        t_ = t
      else:
        t_ = self.words[w_].tag
        
      increment_transition_counts( tc, t_, t )
      _w = self.tokens[i+1]
      if _w != w:
        
        if _w == LexiconState.BOUNDARY_TOKEN:
          _t = 0
        else:
          _t = self.words[_w].tag
          
        increment_transition_counts( tc, t, _t )
    # for
    log_prob = 0.0
    for t_,cc in tc.iteritems():
      d = self.distrs_transition[t_]
      N = self.NUM_TAGS if t_ == 0 else self.NUM_TAGS + 1
      total = d.total
      for t,n in cc.iteritems():
        count = d.counts[t]
        for i in range(n):
          numer = d.alpha + count + i
          denom = d.alpha * N + total + i
          log_prob += math.log( numer / denom )
        total += n
      # for t
    # for t_
    return log_prob

  # returns {None,True,False}
  def calculate_pairwise_suffix_match_(self, w,spans,stem_index, neighbor):
    assert w != neighbor
    if neighbor not in self.words:
      return None
    else:
      w_suffix = WordState.get_last_suffix_static(w,spans,stem_index)
      n_suffix = self.words[neighbor].get_last_suffix()
      m = max( len(w_suffix), len(n_suffix) )
      if w[-m:] == neighbor[-m:]:
        # compare only when they have the same endings
        return w_suffix == n_suffix
      else:
        return None

  def calculate_pairwise_suffix_match(self, w,spans,stem_index, neighbor):
    assert w != neighbor
    if neighbor not in self.words:
      return None
    else:
      w_suffix = WordState.get_last_suffix_static(w,spans,stem_index)
      n_suffix = self.words[neighbor].get_last_suffix()
      m = max( len(w_suffix), len(n_suffix) )
      if m > 0:
        if w[-m:] == neighbor[-m:]:
          # compare only when they have the same endings
          return w_suffix == n_suffix
        else:
          return None
      else:
        return None

  # returns {None,True,False}
  def calculate_pairwise_prefix_match(self, w,spans,stem_index, neighbor):
    assert w != neighbor
    if neighbor not in self.words:
      return None
    else:
      w_prefix = WordState.get_first_prefix_static(w,spans,stem_index)
      n_prefix = self.words[neighbor].get_first_prefix()
      m = max( len(w_prefix), len(n_prefix) )
      if m > 0:
        if w[:m] == neighbor[:m]:
          # compare only when they have the same starting sequence
          return w_prefix == n_prefix
        else:
          return None
      else:
        return None

  def log_uprob_of_new_word_state_fast(self, w,freq,all_possible_tags,spans,stem_index,log_tag_probs,verbose=False):
    for t in all_possible_tags: assert t > 0

    if stem_index > LexiconState.MAX_AFFIX \
       or len(spans) - stem_index > LexiconState.MAX_AFFIX+1:
      return [ -float('inf') ] * len(all_possible_tags)

    if self.USE_STEM_LENGTH_CONSTRAINT:
      #if stem_length(spans,stem_index) <= 1:
      if stem_length_violation(spans,stem_index):
        return [ -float('inf') ] * len(all_possible_tags)

    assert len(log_tag_probs) == len(all_possible_tags)

    segs = [ w[start:stop] for (start,stop) in spans ]
    unique_segs = set(segs)
    new_unique_segs = [ s for s in unique_segs if self.counter_seg[s] == 0 ]
    min_num_segs = len(self.counter_seg) + len(new_unique_segs) # NOTE: we delete keys with zero counts in observe_word_state

    affix_counter = collections.Counter()
    x_new_unique_segs = collections.defaultdict( set )
    for i,s in enumerate(segs):
      affix = WordState.get_seg_affix(i, stem_index)
      affix_counter[affix] += 1
      if self.SEPARATE_LEXICON_SIZE:
        if self.x_counter_seg[affix][s] == 0: x_new_unique_segs[affix].add(s)

    log_prob_lexicon =  0.0;
    if self.SEPARATE_LEXICON_SIZE:
      # TODO: separate size for prefix, stem, suffix
      for affix in WordState.POSSIBLE_AFFIXES:
        #min_num_segs = len([v for v in self.x_counter_seg[affix].values() if v > 0]) + len(x_new_unique_segs[affix])
        min_num_segs = len(self.x_counter_seg[affix]) + len(x_new_unique_segs[affix])
        log_prob_lexicon += log_geometric( self.X_GAMMA_NUM_UNIQUE_SEGS[affix], min_num_segs )
        if verbose:
          print "*lex size(%s) %5.3f" % (affix, log_geometric( self.X_GAMMA_NUM_UNIQUE_SEGS[affix], min_num_segs ))
    else:
      log_prob_lexicon += self.log_prob_seg_lexicon_size( min_num_segs )

    log_prob_lexicon += sum(self.log_prob_segment_length( len(seg) ) for seg in new_unique_segs)
    if verbose:
      vvv = [ self.log_prob_segment_length( len(seg) ) for seg in new_unique_segs ]
      print "*length of new segs %s %5.3f" % (str(vvv),sum(vvv))

    if self.VALENCE_MODEL == 0:
      #log_prob_lexicon += LexiconState.log_prob_num_segments_in_word( len(w), len(spans) )
      log_prob_lexicon += self.LOG_GEOM_TRUNC_NUM_SEGS_PER_WORD.logprob( len(spans) )
    elif self.VALENCE_MODEL == 1:
      log_prob_lexicon += sum(self.LOG_GEOM_TRUNC_NUM_AFFIX_PER_WORD.logprob(affix_counter[x])
                              for x in ['prefix','suffix'])
    else: assert False

    log_probs = [ log_prob_lexicon + v for v in log_tag_probs ]

    # surface forms
    affix_min_vocab_size = self.get_min_affix_vocab_size(segs, stem_index)
    for i,s in enumerate(segs):
      affix = WordState.get_seg_affix(i, stem_index)
      fast_distrs_seg_gt = self.x_fast_distrs_seg_gt[affix]
      Ns = affix_min_vocab_size[affix]

      if affix == 'stem':
        v = fast_distrs_seg_gt.log_probs_no_tag(s, Ns)
        #v2 = math.log( ( self.ALPHA_SEG + self.x_counter_seg[affix][s] ) / ( self.ALPHA_SEG*Ns + sum(self.x_counter_seg[affix].values()) ) )
        #assert v == v2
        for j in range(len(log_probs)): log_probs[j] += v
      else:
        p = fast_distrs_seg_gt.log_probs(s, Ns, all_possible_tags)
        if not (i+1 == stem_index or i == len(segs)-1):
          for t in all_possible_tags: fast_distrs_seg_gt.observe(t,s,1)
        assert len(p) == len(all_possible_tags)
        for j,v in enumerate(p): log_probs[j] += v
          
    # undo cascading counts
    for i,s in enumerate(segs):
      affix = WordState.get_seg_affix(i, stem_index)
      fast_distrs_seg_gt = self.x_fast_distrs_seg_gt[affix]
      if i != stem_index:
        if not (i+1 == stem_index or i == len(segs)-1):
          for t in all_possible_tags: fast_distrs_seg_gt.observe(t,s,-1)
            
    # token-level pairwise suffix
    if self.tokens and ( self.USE_PAIRWISE_SUFFIXES or self.USE_PAIRWISE_PREFIXES):
      log_prob_pairwise_suffix = 0.0
      log_prob_pairwise_prefix = 0.0
      for i in self.word_locations[w]:
        w_ = self.tokens[i-1]
        if w != w_:
          if self.USE_PAIRWISE_SUFFIXES:
            suffix_match = self.calculate_pairwise_suffix_match(w,spans,stem_index, w_)
            log_prob_pairwise_suffix += self.PAIRWISE_SUFFIX_LOGPROB[suffix_match]
          if self.USE_PAIRWISE_PREFIXES:
            prefix_match = self.calculate_pairwise_prefix_match(w,spans,stem_index, w_)
            #if prefix_match: pdb.set_trace()
            log_prob_pairwise_prefix += self.PAIRWISE_PREFIX_LOGPROB[prefix_match]
        else:
          if self.USE_PAIRWISE_SUFFIXES:
            log_prob_pairwise_suffix += self.PAIRWISE_SUFFIX_LOGPROB[True]
          if self.USE_PAIRWISE_PREFIXES:
            log_prob_pairwise_prefix += self.PAIRWISE_PREFIX_LOGPROB[True]
        _w = self.tokens[i+1]
        if w != _w:
          if self.USE_PAIRWISE_SUFFIXES:
            suffix_match = self.calculate_pairwise_suffix_match(w,spans,stem_index, _w)
            log_prob_pairwise_suffix += self.PAIRWISE_SUFFIX_LOGPROB[suffix_match]
          if self.USE_PAIRWISE_PREFIXES:
            prefix_match = self.calculate_pairwise_prefix_match(w,spans,stem_index, _w)
            log_prob_pairwise_prefix += self.PAIRWISE_PREFIX_LOGPROB[prefix_match]

      for j in range(len(log_probs)): log_probs[j] += log_prob_pairwise_suffix
      for j in range(len(log_probs)): log_probs[j] += log_prob_pairwise_prefix

    return log_probs

  def log_prob_seg_lexicon_size(self,k):
    return log_geometric(self.GAMMA_NUM_UNIQUE_SEGS, k-1)

  def calculate_log_tag_probs(self, w, all_possible_tags):
    log_probs = self.get_log_tag_prior(all_possible_tags)
    if self.tokens:
      log_emission_probs = self.log_token_emission(w,all_possible_tags)
      log_trans_probs = [ self.log_token_trans(w,t) for t in all_possible_tags ]
      #for t,p1,p2 in zip(all_possible_tags,log_emission_probs,log_trans_probs): print t,p1,p2
      #pdb.set_trace()
      assert len(log_probs) == len(log_emission_probs)
      assert len(log_probs) == len(log_trans_probs)
      for j,(ep,tp) in enumerate(zip(log_emission_probs,log_trans_probs)):
        log_probs[j] += ep
        log_probs[j] += tp
    return log_probs

    return log_tag_probs

  def get_log_token_emissions(self, w, all_possible_tags):
    assert w not in self.fast_distrs_emit_gt.counts
    Ns = [ 1+len(self.t_counter_token[t]) for t in all_possible_tags ]
    return self.fast_distrs_emit_gt.log_probs_with_Ns(w, Ns, all_possible_tags)

  def get_tag(self, w):
    if not w: return 0
    else: return self.words[w].tag

  def log_prob_token_transitions(self, w, t):
    v = 0.0
    K = self.NUM_TAGS
    for i in self.WLOCS[w]:
      w_ = self.TOKENS[i-1]
      t_ = self.get_tag(w_) if w_ != w else t
      v = v + self.t_fdistr_trans[t_].log_prob_with_N(t, K)

      _w = self.TOKENS[i+1]
      _t = self.get_tag(_w) if _w != w else t
      d = self.t_fdistr_trans[t]
      denom_corr = 1 if t_ == t else 0
      denom = 0.0 + d.alpha * K + d.total + denom_corr
      n = d.counts[_t] if _t in d.counts else 0
      numer_corr = 1 if t_ == t == _t else 0
      numer = 0.0 + d.alpha + n + numer_corr
      v = v + math.log(numer / denom)
    return v

def affix_conflict(stem, affixes, is_prefix):
  for affix in affixes:
    if is_prefix:
      if stem.startswith(affix): return True
    else:
      if stem.endswith(affix): return True
  return False

def stem_length(spans,stem_index):
  (start,stop) = spans[stem_index]
  return stop-start

def stem_length_violation(spans,stem_index):
  assert 0 <= stem_index < len(spans)
  (start,stop) = spans[stem_index]
  stem_len = stop - start
  for i,sp in enumerate(spans):
    if i == stem_index: continue
    (start,stop) = spans[i]
    if stop-start >= stem_len: return True
  return False

def log_geometric(gamma, k):
  assert k >= 0
  return 1.0 * k * math.log(1.0-gamma) + math.log(gamma)

def log_geometric_truncated(gamma,k,N):
  assert 0 <= k <= N
  logp = [ 1.0 * i * math.log(1.0-gamma) + math.log(gamma)
           for i in range(N+1) ]
  b = max(logp)
  logz = math.log( sum( math.exp(i - b) for i in logp ) ) + b
  return logp[k] - logz
    
class LogGeometricTruncated:
  def __init__(self,gamma,N):
    logprobs = [ log_geometric(gamma, i) for i in range(N+1) ]
    b = max(logprobs)
    logz = math.log(sum(math.exp(v-b) for v in logprobs)) + b
    self.logprobs = tuple([ v-logz for v in logprobs ])
    self.N = N

  def logprob(self,k):
    assert 0 <= k <= self.N
    return self.logprobs[k]

class WordState:
  POSSIBLE_AFFIXES = ['stem','prefix','suffix']

  @staticmethod
  def get_seg_affix(i,stem_index):
    if i < stem_index:
      return 'prefix'
    elif i > stem_index:
      return 'suffix'
    else:
      return 'stem'
  
  def seg_affix(self,i):
    return WordState.get_seg_affix(i, self.stem_index)

  @staticmethod
  def binfreq(f):
    return int(math.floor(math.log(f))) if f > 0 else 0

  def __init__(self,w,f,t,s,i,to_seg,to_stem,to_tag):
    assert len(w) > 0
    self.word = w
    assert type(f) == int
    self.freq = f
    self.fbin = WordState.binfreq(f)
    assert t > 0
    self.tag = t
    assert 0 <= i < len(s)
    self.stem_index = i
    assert type(s) == tuple
    self.spans = s # copy.deepcopy(s) # SLOW
    assert type(to_seg) == bool
    self.to_segment = to_seg
    assert type(to_stem) == bool
    self.to_stem = to_stem
    # if fix stem, then must also fix segmentations
    if not to_stem: assert not to_seg
    assert type(to_tag) == bool
    if not to_tag: assert not to_stem
    self.to_tag = to_tag
    
  def __eq__(self,other):
    return self.word == other.word \
           and self.freq == other.freq \
           and self.fbin == other.fbin \
           and self.tag == other.tag \
           and self.spans == other.spans \
           and self.stem_index == other.stem_index \
           and self.to_segment == other.to_segment

  def segments(self):
    segs = []
    for (start,stop) in self.spans:
      segs.append( self.word[start:stop] )
    return segs

  @staticmethod
  def get_last_suffix_static(w,spans,stem_index):
    assert spans[0][0] == 0
    assert spans[-1][1] == len(w)
    if stem_index < len(spans) -1:
      (start,stop) = spans[-1]
      return w[start:stop]
    else:
      return ""

  def get_last_suffix(self):
    return WordState.get_last_suffix_static(self.word,self.spans,self.stem_index)

  @staticmethod
  def get_first_prefix_static(w,spans,stem_index):
    assert spans[0][0] == 0
    assert spans[-1][1] == len(w)
    if stem_index > 0:
      (start,stop) = spans[0]
      return w[start:stop]
    else:
      return ""
    
  def get_first_prefix(self):
    return WordState.get_first_prefix_static(self.word,self.spans,self.stem_index)
  
  def stem(self):
    (start,stop) = self.spans[self.stem_index]
    return self.word[start:stop]

  def __str__(self):
    s = self.segments()
    i = self.stem_index
    flag = "+" if self.to_segment else " "
    return "{" + flag + str(self.tag) + " " + str(len(self.segments())) + " : " \
           + "-".join(s[0:i]) + "[" + s[i] + "]" + "-".join(s[i+1::]) \
           + " " + str(self.freq) + " / " + str(self.fbin) + "}"

  def log_prob_normalized_num_segments(self):
    assert self.to_segment
    num_segs = len(self.spans)
    word_length = len(self.word)
    k = 1.0 * num_segs
    return log_geometric(self.GAMMA_NORMALIZED_NUM_SEGS_PER_WORD, k)

def log_exponential(gamma, k):
  assert gamma > 0
  assert k > 0
  return -1.0 * gamma * math.log(k)

def seg_to_spans(seg):
  i = 0
  spans = []
  for s in seg:
    n = len(s)
    start = i
    stop = i + n
    spans.append( (start,stop) )
    i += n
  return spans


def find_stem_index(tags, t):
  for i,tag in enumerate(tags):
    if tag == t:
      return i
    else:
      a = set(tag.split("_"))
      if t in a:
        return i
  raise Exception(t + " " + " ".join(tags))

def check_gold(gold):
  d = {}
  for sent in gold:
    for segs,tags,flag in sent:
      w = "".join(segs)
      if w not in d: d[w] = {}
      if flag not in d[w]: d[w][flag] = 0
      d[w][flag] = d[w][flag] + 1
  for w,fc in d.iteritems():
    if len(fc) > 1:
      pdb.set_trace()

def calculate_stem_accuracy(gold,state):
  import poke_state
  type_correct = 0.0
  type_total = 0.0
  d_correct = collections.Counter()
  d_total = collections.Counter()
  for phrase in gold:
    for sw in phrase.split():
      segs = sw.split('-')
      word = "".join(segs)
      si = poke_state.pick_stem_heuristically(segs,first=True)
      gold_stem = segs[si]
      predicted_stem = state.words[word].stem()
      if gold_stem == predicted_stem:
        type_correct += 1.0
        d_correct[gold_stem] += 1.0
      type_total += 1.0
      d_total[gold_stem] += 1.0

  e_total = collections.Counter()
  e_correct = collections.Counter()
  for stem,total in d_total.iteritems():
     e_total[ total ] += total
     e_correct[ total ] += d_correct[stem]

  def binf(f):
    if 0 < f <= 10:
      return "0=(0,10]"
    elif 10 < f <= 20:
      return "1=(10,20]"
    elif 20 < f <= 50:
      return "2=(10,50]"
    else:
      return "3=>50"
    
  accuracy_by_stemfreq = {}
  for freq,total in e_total.iteritems():
    correct = e_correct[freq]
    accuracy_by_stemfreq[ binf(freq) ] = 1.0 * correct / total
  type_accuracy = type_correct / type_total;
  return (type_accuracy,accuracy_by_stemfreq)
      
def print_stats(it, state, counts, gold, new_line):
  assert len(counts) == len(gold)

  # use only words with non-zero counts
  gold = tuple([ phrase for phrase,count in zip(gold,counts) if count > 0 ])
  plex = state.to_segmented_lexicon()
  prediction = sb08_data_helper.segment_gold_phrases_type_level(gold, plex)
  [r,p,f1] = sb08_data_helper.score(gold, prediction)
  
  #check_gold(gold)

  segl = state.seg_lens()
  affix_segl = state.get_affix_segment_lengths()
  # pacify scip.stats.scoreatpercentile
  affix_segl = dict([ (k,affix_segl[k] if affix_segl[k] else [0]) for k in ["prefix","stem","suffix"] ])
  prel = affix_segl['prefix']
  sufl = affix_segl['suffix']
  steml = affix_segl['stem']
  
  npref = state.num_prefixes_per_word()
  nsuff = state.num_suffixes_per_word()
  
  #nns = state.normalized_num_seg()
  nns = state.num_seg()
  num_prefix_types = state.num_prefix_types()
  num_suffix_types = state.num_suffix_types()
  num_stem_types = len(set(state.stems()))

  tp = [(1.0*state.distr_type_tag.counts[t]/len(state.words) if t in state.distr_type_tag.counts else 0) for t in range(1,state.NUM_TAGS+1)]
  t = " ".join([ "%3.2f" % v if v > 0.01 else "." for v in tp ])
  logging.info(" it %d r %5.3f p %5.3f f1 %5.3f" % (it,r,p,f1) )
  logging.info("     type_tags %s lex_size = %d" % (t,len(state.counter_seg)))
  logging.info("     SEG_LEN(mean,median,90p) %3.1f %d %d < %3.1f %d %d * %3.1f %d %d > %3.1f %d %d"
               % (
                 numpy.mean(segl), numpy.median(segl), scipy.stats.scoreatpercentile(segl,90),
                 numpy.mean(prel), numpy.median(prel), scipy.stats.scoreatpercentile(prel,90),
                 numpy.mean(steml),numpy.median(steml),scipy.stats.scoreatpercentile(steml,90),
                 numpy.mean(sufl), numpy.median(sufl), scipy.stats.scoreatpercentile(sufl,90)
                 )
               )
  logging.info("     NUM_SEGS %3.1f %d %d < (%d) %3.1f %d %d * (%d) > (%d) %3.1f %d %d"
              % (
                 numpy.mean(nns),  numpy.median(nns),  scipy.stats.scoreatpercentile(nns,90),
                 num_prefix_types,
                 numpy.mean(npref),numpy.median(npref),scipy.stats.scoreatpercentile(npref,90),
                 num_stem_types,
                 num_suffix_types,
                 numpy.mean(nsuff),numpy.median(nsuff),scipy.stats.scoreatpercentile(nsuff,90)
                 ))
  logging.info("     top PREFIXES (%d): %s" % (sum(state.x_counter_seg['prefix'].values()), str(state.x_counter_seg['prefix'].most_common(10))))
  logging.info("     top SUFFIXES (%d): %s" % (sum(state.x_counter_seg['suffix'].values()), str(state.x_counter_seg['suffix'].most_common(10))))
  logging.info("     top stems (%d): %s" % (sum(state.x_counter_seg['stem'].values()), str(state.x_counter_seg['stem'].most_common(5))))

class Worker:
  def __init__(self, i, func, master_q):
    self.id = i
    self.q = multiprocessing.Queue(1)
    self.p = multiprocessing.Process(target=func, args=[self.q, master_q])
    self.p.start()
  def do(self, cmd, arg):
    item = (cmd,arg)
    self.q.put( item, True)
  def add_word(self, ws):
    #self.do( "add-word", ws )
    self.q.put( ("add-word",ws) )
  def remove_word(self,w):
    #self.do( "remove-word", w)
    self.q.put( ("remove-word", w) )

def worker_func(inq, outq):
  state = None
  log_tag_probs = None
  while True:
    (task, param) = inq.get(True) # block=True
    if task == "kill":
      break
    elif task == "init-state":
      state = param
    elif task == "remove-word":
      state.remove_word( param )
    elif task == "add-word":
      #new_ws = WordState(w,freq,chosen_tag,chosen_spans,chosen_stem_index,flag)
      state.add_word_state( param )
    elif task == "set-log-tag-probs":
      log_tag_probs = param
    elif task == "compute-list":
      res = []
      (subspace,all_possible_tags) = param
      for (i,w,freq,s,j) in subspace:
        p = state.log_uprob_of_new_word_state_fast(w,freq,all_possible_tags,s,j,log_tag_probs)
        res.append( (i,p,s,j) )
      outq.put_nowait(res)
    elif task == "sample-list":
      u = []
      (subspace,all_possible_tags) = param
      for (i,w,freq,s,j) in subspace:
        p = state.log_uprob_of_new_word_state_fast(w,freq,all_possible_tags,s,j,log_tag_probs)
        u.append( (i,p,s,j) )
      nt = len(all_possible_tags)
      temp2 = [ zip([i]*nt, p, [s]*nt, all_possible_tags, [j]*nt ) for (i,p,s,j) in u ]
      #print len(subspace),len(temp2)
      (ii,pp,ss,tt,jj) = zip(*[ hypothesis for sublist in temp2 for hypothesis in sublist ])
      (ind,prob,logz) = sample_from_unnormalized_logs(pp)
      iii = ii[ind]
      sss = ss[ind]
      ttt = tt[ind]
      #print ttt,tt
      #print all_possible_tags
      jjj = jj[ind]
      res = [ (iii,logz,sss,ttt,jjj) ]
      outq.put_nowait(res)
    else:
      assert False

def count_word_types(phrase_counts, phrases):
  def index_word(d, w):
    if w not in d: d[w] = len(d) + 1
  
  wc = collections.Counter()
  w2i = {}
  assert len(phrase_counts) == len(phrases)

  for count,p in zip(phrase_counts,phrases):
    if count < 0: continue
    to_segment = count > 0
    words = p.split()
    # no loop if p is empty (i.e. boundary)
    for w in words:
      index_word(w2i, w)
      #wc[w] += count
      #wc[w] += 1
      wc[w] += 1 if to_segment else 0
      #wc[w] += 1 if w in words_to_segment else 0

  # includes words with count zero 
  #word_counts = tuple([ (w,c) for w,c in sorted( wc.iteritems(), key=lambda (w,c): w2i[w] ) ])
  word_counts = tuple([ (w,c) for w,c in sorted( wc.iteritems(), key=lambda (w,c): w) ])
  return word_counts

class MultiprocessingPool:
  WORKER_SAMPLE_LIST = False

  def __init__(self, state, num_workers):
    self.master_q = multiprocessing.Queue(num_workers)
    self.workers = [ Worker(i, worker_func, self.master_q) for i in range(num_workers) ]
    state_copy = copy.deepcopy(state)
    for worker in self.workers: worker.do( "init-state", state_copy )

  def remove_word(self,w):
    for worker in self.workers: worker.remove_word(w) #worker.do( "remove-word", w )

  def compute_sampling_logprobs(self, w,freq, log_tag_probs, all_possible_spans, all_possible_tags, FIX_STEM, stem_index):
    for worker in self.workers: worker.do( "set-log-tag-probs", log_tag_probs )

    if FIX_STEM:
      assert len(all_possible_spans) == 1
      state_space = [ (s,j)
                      for s in all_possible_spans
                      for j in [ stem_index ] ]
    else:
      state_space = [ (s,j)
                      for s in all_possible_spans
                      for j in range(len(s)) ]
    state_space = [ (i,w,freq,s,j) for i,(s,j) in enumerate(state_space) ]
    #random.shuffle(state_space)
    NUM_PROCESSES = len(self.workers)
    for i,subspace in enumerate(list(chunk(state_space, NUM_PROCESSES))):
      if MultiprocessingPool.WORKER_SAMPLE_LIST:
        self.workers[i].do( "sample-list", (subspace,all_possible_tags) )
      else:
        self.workers[i].do( "compute-list", (subspace,all_possible_tags) )

    count = 0
    results = []
    while count != NUM_PROCESSES:
      results.extend( self.master_q.get() )
      count = count + 1
    results = sorted(results,key=itemgetter(0))
    if MultiprocessingPool.WORKER_SAMPLE_LIST:
      (ii,log_unnormalized_posterior,spans,tags,stem_indices) = zip(*results)
    else:
      temp = [ r[1::] for r in results ]
      nt = len(all_possible_tags)
      temp2 = [ zip( p, [s]*nt, all_possible_tags, [j]*nt ) for (p,s,j) in temp ]
      (log_unnormalized_posterior,spans,tags,stem_indices) = zip(*[ hypothesis for sublist in temp2 for hypothesis in sublist ])
      return (log_unnormalized_posterior,spans,tags,stem_indices)

  def add_word_state(self,new_ws):
    for worker in self.workers: worker.add_word(new_ws)

  def killall(self):
    for worker in self.workers: worker.do( "kill", None )

def save_last_n_history(history, it, state, n):
  assert type(it) == int
  if len(history) < n:
    history[it] = copy.deepcopy(state.words)
  elif len(history) == n:
    k = min(history.keys())
    del history[k]
    assert it > max(history.keys())
    history[it] = copy.deepcopy(state.words)
  else:
    assert False

def get_training_data(gold,flags):
  assert len(gold) == len(flags)
  d = collections.defaultdict( collections.Counter )
  for gphrase,f in zip(gold,flags):
    if not f: continue
    for gword in gphrase.split():
      w = gword.replace('-','')
      d[w][gword] += 1
  e = {}
  for w,counter in d.iteritems():
    sorted_counts = sorted( counter.iteritems(), key=itemgetter(1), reverse=True)
    gold_word = sorted_counts[0][0]
    gold_segs = gold_word.split('-')
    gold_spans = seg_to_spans(gold_segs)
    e[w] = tuple(gold_spans)
  return e

def run_gibbs(hyperparameters, phrase_counts, phrases, \
              gold, gold_tags, \
              seq, use_seq_suffix, use_seq_prefix, \
              NUM_PROCESSES, numit, num_tags, FIX_TAGS, FIX_SEGS, FIX_STEM, \
              init_tag, init_seg, init_stem, \
              state0, outfile, sep_lex_size, flags):
  
  USE_MULTIPROCESSING = NUM_PROCESSES > 0
  #NUM_PROCESSES = 4

  logging.info("multiprocessing %s %d" % (USE_MULTIPROCESSING, NUM_PROCESSES))
  wordcounts = count_word_types(phrase_counts, phrases)
  print "to_segment words =",len([w for w,c in wordcounts if c > 0 ])

  (post_lexicon,tag_mapping) = (None,None)

  if init_tag == "state":
    num_tags = max( max([ws.tag for w,ws in state0.words.iteritems()]), num_tags)
    logging.info("init_tag == state. setting num_tags = %d" % num_tags)

  if flags: training_word_spans = get_training_data(gold,flags)

  seq_data = phrases if seq else None
  state = LexiconState(hyperparameters, wordcounts,seq_data,gold,post_lexicon, init_tag, init_seg, init_stem, num_tags, state0, sep_lex_size, \
                       use_seq_suffix, use_seq_prefix)
  assert state.validate()
  print_stats(-1, state, phrase_counts,gold, True)

  if USE_MULTIPROCESSING: pool = MultiprocessingPool(state, NUM_PROCESSES)
  start_dump = time.time()
  history = {}

  for it in range(numit):

    ops = {}
    wordcounts_shuffled = list(wordcounts)
    random.shuffle(wordcounts_shuffled)
    for (w,freq) in wordcounts_shuffled:
      old_ws = state.remove_word(w)
      to_segment = old_ws.to_segment
      to_stem = old_ws.to_stem
      to_tag = old_ws.to_tag
      #assert old_ws.to_segment
      if USE_MULTIPROCESSING: pool.remove_word(w)

      if FIX_SEGS or not to_segment:
        all_possible_spans = tuple([ old_ws.spans ])
      else:
        if flags and w in training_word_spans:
          all_possible_spans = tuple([ training_word_spans[w] ])
        else:
          all_possible_spans = []
          for k in range(1,LexiconState.MAX_SPANS+1): all_possible_spans.extend( spans_of_length(len(w),k) )
          all_possible_spans = tuple([ tuple(s) for s in all_possible_spans ])

      if FIX_TAGS or not to_tag:
        all_possible_tags = tuple([ old_ws.tag ])
      else:
        all_possible_tags = tuple(range(1,state.NUM_TAGS+1))

      log_tag_probs = state.calculate_log_tag_probs(w, all_possible_tags) if len(all_possible_tags) > 1 else [0.0] * len(all_possible_tags)
        
      if USE_MULTIPROCESSING and len(all_possible_spans) >= NUM_PROCESSES:
        if FIX_STEM or not to_stem:
          assert FIX_SEGS or not to_segment
          stem_index = old_ws.stem_index
          fix_word_stem = True
        else:
          stem_index = None
          fix_word_stem = False
        (log_unnormalized_posterior,spans,tags,stem_indices) = pool.compute_sampling_logprobs(w,freq, log_tag_probs, all_possible_spans, all_possible_tags, fix_word_stem, stem_index)
      else:
        if FIX_STEM or not to_stem:
          assert FIX_SEGS or not to_segment
          assert len(all_possible_spans) == 1
          temp = [ ( state.log_uprob_of_new_word_state_fast(w,freq,all_possible_tags,s,j,log_tag_probs), \
                     s, \
                     j  \
                     )
                   for s in all_possible_spans
                   for j in [ old_ws.stem_index ]
                   ]
        else:
          temp = [ ( state.log_uprob_of_new_word_state_fast(w,freq,all_possible_tags,s,j,log_tag_probs), \
                     s, \
                     j  \
                   )
                   for s in all_possible_spans
                   for j in range(len(s)) ]
        nt = len(all_possible_tags)
        temp2 = [ zip( p, [s]*nt, all_possible_tags, [j]*nt ) for (p,s,j) in temp ]
        (log_unnormalized_posterior,spans,tags,stem_indices) = zip(*[ hypothesis for sublist in temp2 for hypothesis in sublist ])

      (ind,prob,logz) = sample_from_unnormalized_logs( log_unnormalized_posterior )
      assert ind >= 0
      chosen_spans = spans[ind]
      chosen_tag = tags[ind]
      chosen_stem_index = stem_indices[ind]
      new_ws = WordState(w,freq,chosen_tag,chosen_spans,chosen_stem_index,old_ws.to_segment,old_ws.to_stem,old_ws.to_tag)
      state.add_word_state(new_ws)

      if USE_MULTIPROCESSING: pool.add_word_state(new_ws)

      op = op_type(new_ws, old_ws)
      if op not in ops: ops[op] = 0
      ops[op] = ops[op] + 1
      #assert state.validate()
    # for w,f
    state.iter = it
    save_last_n_history(history, it, state, 2)
    end_dump = time.time()
    time_last_dump = end_dump - start_dump
    print_stats(it, state, phrase_counts, gold, True)
    #logging.info("     ops = %s" % ops)
    optstr = ""
    for optype,opcount in sorted(ops.iteritems(), key=itemgetter(1), reverse=True):
      optstr += "%s:%d " % (optype,opcount)
    logging.info("     ops = %s" % optstr)
    assert state.validate()
    #pdb.set_trace()
    if outfile and time_last_dump > 60:
      of = open(outfile, 'wb')
      cPickle.dump(state, of, -1)
      of.close()
      start_dump = time.time()
  # for it
  if USE_MULTIPROCESSING: pool.killall()
  state.history = history
  return state

def print_sampling_probs(w,log_unnormalized_posterior,spans,tags,stem_indices,logz):
  def format_seg(w,span,stem_index):
    prefix = "-".join([ w[start:stop] for i,(start,stop) in enumerate(span) if i < stem_index ])
    (start,stop) = span[stem_index]
    stem = "[" + w[start:stop] + "]"
    suffix = "-".join([ w[start:stop] for i,(start,stop) in enumerate(span) if i > stem_index ])
    return prefix + stem + suffix
    
  for i,(logp,span,tag,stem_index) in enumerate(zip(log_unnormalized_posterior,spans,tags,stem_indices)):
    if logp > float('-inf'):
      print "%d %5.3f %d %s" % (i, logp-logz, tag, format_seg(w,span,stem_index) )

def chunk(l, n):
  N = len(l)
  p = int(N/n)
  q = N % n
  z = [p] * n
  for i in range(q): z[i] = z[i] + 1
  z.insert(0,0)
  for i in range(1,n+1): z[i] = z[i] + z[i-1]
  #print "z=",z
  for i in range(n): yield l[ z[i]: z[i+1] ]

def op_type(new_ws, old_ws):
  tag_change = "*" if new_ws.tag != old_ws.tag else ""
  stem_change = ""
  if new_ws.stem_index > old_ws.stem_index:
    stem_change = ">"
  elif new_ws.stem_index < old_ws.stem_index:
    stem_change = "<"
  changes = tag_change + stem_change
  if len(new_ws.spans) > len(old_ws.spans):
    return changes + "split"
  elif len(new_ws.spans) < len(old_ws.spans):
    return changes + "merge"
  elif new_ws.spans == old_ws.spans and not changes:
    return "noop"
  else: return changes + "shuffle"
  
def sample_from_unnormalized_logs(vals):
  if len(vals) == 1: return 0,1.0,vals[0]
  
  b = max(vals)
  logz = math.log( sum([ math.exp(v-b) for v in vals ]) ) + b
  probs = [ math.exp(v-logz) for v in vals ]
  if not math.fabs(sum(probs) - 1.0) < 1e-10: print sum(probs),vals
  assert math.fabs(sum(probs) - 1.0) < 1e-10
  random_number = random.random()
  p = 0.0
  for i,v in enumerate(probs):
    p = p + v
    if random_number < p:
      return i,probs[i],logz
  assert False
  
def spans_of_length(n,k):
  if k > n: return []
  if k == 1: return [[(0,n)]]
  ret = []
  for split in range(1,n):
    for span in spans_of_length(n-split,k-1):
      shifted = [(s+split,t+split) for (s,t) in span]
      shifted.append((0,split))
      ret.append(sorted(shifted, key=itemgetter(0)))
  return ret

def create_options_parser():
  usage = "usage: %prog -h/--help"
  p = optparse.OptionParser(usage)
  p.add_option("-g","--gold",   dest="gold",    help="GOLD segmentation filename (.fpg)")
  p.add_option("-q","--seq",    dest="seq",     help=".fpg has SEQ information", default=False, action="store_true")
  p.add_option("-s","--suffix", dest="suffix",  help="use seq suffix (enables -q)", default=False, action="store_true")
  p.add_option("-p","--prefix", dest="prefix",  help="use seq prefix (enables -q)", default=False, action="store_true")
  p.add_option("-i","--iters",  dest="numit",   help="number of ITERS", type="int", default=50)
  p.add_option("-t","--tags",   dest="numtags", help="number of TAGS",  type="int", default=1)
  p.add_option("-u","--cpu",  dest="num_processes",   help="number of worker PROCESSES (0 to disable)", type="int", default=0)
  p.add_option("-r","--rseed",  dest="randseed",help="RANDdom SEED (default = 0)",type="int",default=0)
  p.add_option(  "--config",    dest="config",  help="CONFIG filename. Note: command line overwrites config file")
  p.add_option(  "--init-tag",  dest="init_tag",help="INIT TAG (uniform,state)", default="uniform")
  p.add_option(  "--init-seg",  dest="init_seg",help="INIT SEG (whole-word,state)", default="whole-word")
  p.add_option(  "--init-stem", dest="init_stem",help="INIT STEM index (first,state)", default="first")
  p.add_option(  "--state", dest="state", help="STATE used for --init-tag and --init-seg")
  p.add_option(  "--fix-tags", dest="fix_tags", help="FIX TAGS; don't sample them", default=False, action="store_true")
  p.add_option(  "--fix-segs", dest="fix_segs", help="FIX SEGMENTATIONS; don't sample them", default=False, action="store_true")
  p.add_option(  "--fix-stem", dest="fix_stem", help="FIX STEM; don't sample them (implies --fix-segs)", default=False, action="store_true")
  p.add_option(  "--eval", dest="eval", help="EVALuation score of state on gold", default=False, action="store_true")
  p.add_option(  "--cut", dest="cut", help="CUT gold data to n rows", type="int", default=-1)
  p.add_option(  "--sep-lex-size", dest="sep_lex_size", help="seperate lexicon size", default=True, action="store_true")
  p.add_option(  "--flags", dest="flags", help="flags indicating which rows are labeled training data")
  p.add_option("-o","--outfile",dest="outfile", help="pickle OUTFILEname")
  p.add_option("-d","--outdict",dest="outdict", help="output segmentation dictionary filename")
  return p

def get_hyperparameters(filename,section):
  p = ConfigParser.ConfigParser()
  p.read(filename)
  return dict([ (k.upper(), eval(p.get(section,k)) ) for k in p.options(section) ])

def strip_segmentation_from_word(word):
  r = re.compile("^[\d-]+$")
  if r.match(word):
    return word
  else:
    return word.replace('-','')
            
def strip_segmentations(gold_cols, start, end):
  assert type(gold_cols[0]) == tuple
  unlabeled_cols = [ gold_cols[0] ]

  for i,col in enumerate(gold_cols):
    if i == 0: continue
    if i >= start and i <= end:
      #unlabeled_cols.append( tuple([ phrase.replace('-','') for phrase in col ]) )
      unsegmented_col = []
      for phrase in col:
        unsegmented_phrase = []
        for w in phrase.split():
          #if w == "-"
          unsegmented_phrase.append( strip_segmentation_from_word(w) )
        unsegmented_phrase = "\t".join(unsegmented_phrase)
        unsegmented_col.append( unsegmented_phrase )
      unlabeled_cols.append( tuple(unsegmented_col) )
    else:
      assert type(col) == tuple
      unlabeled_cols.append( col )
  return tuple(unlabeled_cols)

  
def run_eval(state, phrase_counts, gold):
  print_stats(-1, state, phrase_counts, gold, True)

def get_gold_type_segs(gold):
  d = collections.defaultdict( collections.Counter )
  for segmented_phrase in gold:
    for segmented_word in segmented_phrase.split():
      segs = segmented_word.split('-') if not LexiconState.FROZEN_RE.match(segmented_word) else tuple([segmented_word])
      w = "".join(segs)
      d[w][segmented_word] += 1
  e = {}
  for w,counts in d.iteritems():
    sorted_counts = sorted(counts.iteritems(), key=itemgetter(1), reverse=True)
    gold_segs = sorted_counts[0][0].split('-')
    e[w] = tuple(gold_segs)
  return e

def load_flags(infile):
  return tuple([ line.strip() for line in infile ])

def main():
  logging.basicConfig(stream=sys.stdout,level=logging.INFO)
  parser = create_options_parser()
  (options,args) = parser.parse_args()
  if options.config:
    i = options.config.find(':')
    filename = options.config[:i]
    section = options.config[i+1:]
    hyperparameters = get_hyperparameters(filename, section)
  else:
    hyperparameters = {}

  gold_cols = sb08_data_helper.read_file( sb08_data_helper.gzopen(options.gold) )
  if options.cut > 0:
    gold_cols = tuple( tuple(col[0: options.cut]) for col in gold_cols )

  unlabeled_cols = strip_segmentations(gold_cols, 1, 1)
  assert gold_cols[0] == unlabeled_cols[0] # freq

  gold_tags = gold_cols[2] if options.init_tag.startswith("gold") else None

  state0 = cPickle.load(open(options.state,'rb')) if options.state else None
  flags = load_flags(open(options.flags)) if options.flags else None
  assert not flags or len(flags) == len(gold_cols[1])

  random.seed( options.randseed )

  if options.eval:
    run_eval(state0, unlabeled_cols[0], gold_cols[1])
  else:
    if options.fix_stem: options.fix_segs = True

    state = run_gibbs(hyperparameters, unlabeled_cols[0], unlabeled_cols[1], \
                      gold_cols[1], gold_tags, \
                      options.seq, options.suffix, options.prefix, \
                      options.num_processes, options.numit, options.numtags, \
                      options.fix_tags, options.fix_segs, options.fix_stem, \
                      options.init_tag, options.init_seg, options.init_stem, state0, \
                      options.outfile, options.sep_lex_size, flags)

  if options.outfile:
    outfile = open(options.outfile, 'wb')
    cPickle.dump(state, outfile, -1)
    outfile.close()
  if options.outdict:
    outdict = open(options.outdict,'w')
    for w,ws in state.words.iteritems():
      print >>outdict,"%s %d %d %s" % (w, len(ws.spans), ws.stem_index, " ".join(ws.segments()))
    outdict.close()

if __name__ == '__main__':
  main()
  #cProfile.run("main()", 'profile-data')
